from typing import Dict, Optional
from datetime import datetime

from config.frozen import Frozen
from config.constants import WEEKDAY
from models.timetable import Timetable
from models.shift import Shift

from xtb import Xtb



class Symbol(Frozen):
    def __init__(self, xtb: Xtb, symbol: str):
        symbol_dict: Dict = xtb.get_client().get_symbol(symbol=symbol)
        self.__symbol: str = symbol_dict['symbol']
        self.__currency: str = symbol_dict['currency']
        self.__currency_profit: str = symbol_dict['currencyProfit']
        self.__contract_size: str = symbol_dict['contractSize']
        self.__precision: str = symbol_dict['precision']
        self.__bid: str = symbol_dict['bid']
        self.__ask: str = symbol_dict['ask']
        self.__swap_long: str = symbol_dict['swapLong']
        self.__swap_short: str = symbol_dict['swapShort']
        self.__spread: float = symbol_dict['spreadRaw']
        self.__timetable: Timetable = xtb.get_symbol_timetable(symbol=symbol)
        
    @property
    def symbol(self) -> str:
        return self.__symbol

    @symbol.setter
    def symbol(self, value: str) -> None:
        self.__symbol = value

    @property
    def currency(self) -> str:
        return self.__currency

    @currency.setter
    def currency(self, value: str) -> None:
        self.__currency = value

    @property
    def currency_profit(self) -> str:
        return self.__currency_profit

    @property
    def contract_size(self) -> str:
        return self.__contract_size

    @contract_size.setter
    def contract_size(self, value: str) -> None:
        self.__contract_size = value

    @property
    def precision(self) -> str:
        return self.__precision

    @precision.setter
    def precision(self, value: str) -> None:
        self.__precision = value

    @property
    def bid(self) -> str:
        return self.__bid

    @bid.setter
    def bid(self, value: str) -> None:
        self.__bid = value

    @property
    def ask(self) -> str:
        return self.__ask

    @ask.setter
    def ask(self, value: str) -> None:
        self.__ask = value

    @property
    def swap_long(self) -> str:
        return self.__swap_long

    @swap_long.setter
    def swap_long(self, value: str) -> None:
        self.__swap_long = value

    @property
    def swap_short(self) -> str:
        return self.__swap_short

    @swap_short.setter
    def swap_short(self, value: str) -> None:
        self.__swap_short = value

    @property
    def timetable(self) -> Timetable:
        return self.__timetable

    @timetable.setter
    def timetable(self, value: Timetable) -> None:
        self.__timetable = value

    @property
    def spread(self) -> float:
        return self.__spread

    @spread.setter
    def spread(self, value: float) -> None:
        self.__spread = value

    def get_active_shift(self) -> Optional[Shift]:
        for shift in getattr(self.timetable, WEEKDAY[datetime.today().isoweekday()]):
            if shift.from_ts < datetime.now() < shift.to_ts:
                return shift



from typing import List, Dict, Optional
from datetime import datetime
from retrying import retry

from config.exception import ArbitragePartiallyClosedException, TradeNotFoundException, \
    retry_if_arbitrage_partially_closed_exception
from config.frozen import Frozen
from config.constants import Cmd
from config.logger import logger
from models.position import Position
from models.symbol import Symbol
from models.shift import Shift
from models.trade import Trade
from utils import get_today_ini, get_today_end
from xtb import Xtb


class Arbitrage(Frozen):
    def __init__(self,
                 xtb: Xtb,
                 name: str,
                 active: bool,
                 tp: float,
                 sl: float,
                 positions: List[Dict]):
        self.__name: str = name
        self.__active: bool = active
        self.__tp: float = tp
        self.__sl: float = sl
        self.__positions: List[Position] = self.__init_positions(xtb=xtb, setup_positions=positions)
        self.__shift: Optional[Shift] = self.__extract_shift()
        self.__status: str = ""
        self.__profit: float = 0
        self.__action: str = ""

    def __init_positions(self, xtb: Xtb, setup_positions: List[Dict]) -> List[Position]:
        positions = []
        for setup_position in setup_positions:
            if "trade" not in setup_position or not setup_position['trade']:
                setup_position['trade'] = dict(
                    cmd=getattr(Cmd(), setup_position['direction']),
                    volume=setup_position['volume'],
                    symbol=setup_position['symbol'],
                    comment=self.__name)
            positions.append(Position(xtb=xtb, **setup_position))
        return positions

    @property
    def name(self) -> str:
        return self.__name

    @property
    def active(self) -> bool:
        return self.__active

    @property
    def active_on_off(self) -> str:
        return "On" if self.__active else "Off"

    @property
    def tp(self) -> float:
        return self.__tp

    @property
    def sl(self) -> float:
        return self.__sl

    @property
    def shift(self) -> Shift:
        return self.__shift

    @property
    def positions(self) -> List[Position]:
        return self.__positions

    @property
    def action(self) -> str:
        return self.__action

    @action.setter
    def action(self, value: str) -> None:
        self.__action = value

    @property
    def status(self) -> str:
        return self.__status

    @status.setter
    def status(self, value: str) -> None:
        self.__status = value

    @property
    def symbols(self) -> List[Symbol]:
        return list(map(lambda position: position.symbol, self.positions))

    @property
    def orders(self) -> List[int]:
        return list(map(lambda position: position.order, self.positions))

    @property
    def trades(self) -> List[Trade]:
        return list(map(lambda position: position.trade, self.positions))

    @property
    def is_working(self):
        return all(map(lambda p: p.is_working, self.positions))

    @property
    def is_completely_accepted(self):
        accepted = list(map(lambda p: p.trade.accepted, self.positions))
        return sum(accepted) == len(accepted)

    @property
    def is_partially_failed(self):
        accepted = list(map(lambda p: p.trade.accepted, self.positions))
        return 0 < sum(accepted) < len(accepted)

    @property
    def is_completely_failed(self):
        accepted = list(map(lambda p: p.trade.accepted, self.positions))
        return 0 == sum(accepted)

    def __extract_shift(self) -> Optional[Shift]:
        if self.is_working:
            shift = Shift(from_ts=get_today_ini(), to_ts=get_today_end())
            for position in self.__positions:
                shift.from_ts = max([shift.from_ts, position.shift.from_ts])
                shift.to_ts = min([shift.to_ts, position.shift.to_ts])
            return shift

    def __get_trades(self, xtb: Xtb):
        trades = xtb.get_client().get_trade_records(orders=self.orders)
        for position in self.positions:
            trade = list(filter(lambda t: t['order'] == position.order, trades))
            if trade:
                position.trade = Trade(trade_raw=list(trade)[0])
            else:
                logger.debug(f"No trade found for order '{position.order} - {position.symbol.symbol}'")
                raise TradeNotFoundException(f"No trade found for order {position.order}")

    def is_operable(self, closing_threshold: int):
        return (self.shift.to_ts - datetime.now()).total_seconds() > closing_threshold

    @property
    def profit(self) -> float:
        return round(sum(map(lambda p: p.profit, self.positions)), 2)

    @retry(retry_on_exception=retry_if_arbitrage_partially_closed_exception,
           stop_max_attempt_number=5, wait_exponential_multiplier=1000, wait_exponential_max=10000)
    def transact(self, xtb: Xtb, _type: int) -> None:
        keep_opening = True
        for idx, position in enumerate(self.positions):
            if keep_opening:
                if not position.trade.accepted:
                    position.trade.transact(xtb=xtb, _type=_type)
                    self.positions[idx].action = position.trade.transaction.status
                if not position.trade.accepted:
                    keep_opening = False

        if self.is_completely_accepted:
            logger.debug("Arbitrage process was SUCCESSFUL and is COMPLETELY CLOSED")
            self.action = "closed"
        if self.is_partially_failed:
            logger.critical("Arbitrage process has FAILED and is PARTIALLY CLOSED")
            self.action = "PARTIALLY closed"
            raise ArbitragePartiallyClosedException("Arbitrage process has FAILED and is PARCIALLY CLOSED")
        if self.is_completely_failed:
            self.action = "COMPLETELY open"
            logger.warning("Arbitrage process has FAILED and is COMPLETELY OPEN")



from typing import Dict, Optional

from config.frozen import Frozen
from config.constants import Cmd
from models.symbol import Symbol
from models.shift import Shift
from models.trade import Trade

from xtb import Xtb


class Position(Frozen):
    def __init__(self,
                 xtb: Xtb,
                 order: int,
                 direction: str,
                 symbol: str,
                 volume: float,
                 trade: Dict):
        self.__direction: int = getattr(Cmd, direction)
        self.__order: int = order
        self.__symbol: Symbol = Symbol(xtb=xtb, symbol=symbol)
        self.__trade: Trade = Trade(trade_raw=trade)
        self.__shift: Optional[Shift] = self.__symbol.get_active_shift()
        self.__volume: float = volume
        self.__action: str = ""

    @property
    def direction(self) -> int:
        return self.__direction

    @property
    def order(self) -> int:
        return self.__order

    @order.setter
    def order(self, value: int) -> None:
        self.__order = value

    @property
    def trade(self) -> Trade:
        return self.__trade

    @trade.setter
    def trade(self, value: Dict) -> None:
        self.__frozen__ = False
        self.__trade = value
        self.__frozen__ = True

    @property
    def symbol(self) -> Symbol:
        return self.__symbol

    @property
    def shift(self) -> Shift:
        return self.__shift

    @shift.setter
    def shift(self, value: Shift) -> None:
        self.__shift = value

    @property
    def volume(self) -> float:
        return self.__volume

    @volume.setter
    def volume(self, value: float) -> None:
        self.__volume = value

    @property
    def is_working(self) -> bool:
        return bool(self.__shift)

    @property
    def action(self) -> str:
        return self.__action

    @action.setter
    def action(self, value: str) -> None:
        self.__action = value

    @property
    def name(self) -> str:
        return f"{self.symbol} {self.direction} {self.volume}"

    @property
    def profit(self) -> float:
        if self.trade:
            return round(self.trade.profit + self.trade.storage, 2)
        return 0



import json
import socket
from time import sleep
import ssl
from random import uniform
from typing import Dict, List
from datetime import datetime
from retrying import retry

from config.constants import API_MAX_CONN_TRIES, API_MIN_SECONDS_BETWEEN_REQUESTS, API_READ_TIMEOUT, \
    XAPI_PORT_DICT, XAPI_ADDRESS, Type
from config.logger import logger
from config.exception import LoginException, GetSymbolException, GetMarginLevel, GetMarginTrade, \
    GetTradingHoursException, GetTradeRecordsException, GetTradeException, NotReceivedDataError, \
    OrderNotSentException, OrderStatusNotSentException, GetTickPricesException, retry_if_get_tick_prices_exception
from utils import unix_ts_to_ts, unix_t_to_ts, ts_to_unix_ts


last_request = datetime.now()


class JsonSocket(object):
    def __init__(self, address, port, encrypt: bool = False):
        self._ssl = encrypt 
        if not self._ssl:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket = ssl.wrap_socket(sock)
        self.conn = self.socket
        self._timeout = None
        self._address = address
        self._port = port
        self._decoder = json.JSONDecoder()
        self._receivedData = ''

    def connect(self):
        for i in range(API_MAX_CONN_TRIES):
            try:
                self.socket.connect((self.address, self.port))
            except socket.error as msg:
                logger.error(f"SockThread Error: {msg}")
                sleep(API_MIN_SECONDS_BETWEEN_REQUESTS)
                continue
            return True
        return False

    def _send_obj(self, obj: Dict):
        self._waiting_send(json.dumps(obj))

    def _waiting_send(self, msg: str):
        if self.socket:
            sent = 0
            msg = msg.encode('utf-8')
            global last_request
            while sent < len(msg):
                sent += self.conn.send(msg[sent:])
                if (datetime.now() - last_request).total_seconds() > API_MIN_SECONDS_BETWEEN_REQUESTS:
                    last_request = datetime.now()
                else:
                    sleep(API_MIN_SECONDS_BETWEEN_REQUESTS)

    def _read(self, bytes_size: int = 4096) -> Dict:
        if not self.socket:
            raise RuntimeError("socket connection broken")
        read_time = 0
        while read_time < API_READ_TIMEOUT:
            char = self.conn.recv(bytes_size).decode()
            self._receivedData += char
            try:
                # Dict, int
                resp, size = self._decoder.raw_decode(self._receivedData)
                if size == len(self._receivedData):
                    self._receivedData = ''
                    break
                elif size < len(self._receivedData):
                    self._receivedData = self._receivedData[size:].strip()
                    break
            except ValueError as e:
                resp = str(e)
                sleep_time = uniform(0.5, 1)
                read_time += sleep_time
                sleep(sleep_time)
                continue
        if read_time > API_READ_TIMEOUT:
            raise NotReceivedDataError("Xtb API is not responding and no data is being received")
        logger.debug(f'Received: {resp}')
        return resp

    def _read_obj(self) -> Dict:
        return self._read()

    def close(self):
        logger.debug("Closing socket")
        self._close_socket()
        if self.socket is not self.conn:
            logger.debug("Closing connection socket")
            self._close_connection()

    def _close_socket(self):
        self.socket.close()

    def _close_connection(self):
        self.conn.close()

    def _get_timeout(self):
        return self._timeout

    def _set_timeout(self, timeout):
        self._timeout = timeout
        self.socket.settimeout(timeout)

    def _get_address(self):
        return self._address

    def _set_address(self, address):
        pass

    def _get_port(self):
        return self._port

    def _set_port(self, port):
        pass

    def _get_encrypt(self):
        return self._ssl

    def _set_encrypt(self, encrypt):
        pass

    timeout = property(_get_timeout, _set_timeout, doc='Get/set the socket timeout')
    address = property(_get_address, _set_address, doc='read only property socket address')
    port = property(_get_port, _set_port, doc='read only property socket port')
    encrypt = property(_get_encrypt, _set_encrypt, doc='read only property socket port')
    
    
class XtbClient(JsonSocket):
    def __init__(self, mode: str):
        super(XtbClient, self).__init__(address=XAPI_ADDRESS, port=XAPI_PORT_DICT[mode], encrypt=True)
        if not self.connect():
            raise Exception(f"Cannot connect to streaming on "
                            f"{XAPI_ADDRESS}: {XAPI_PORT_DICT[mode]} after {API_MAX_CONN_TRIES} retries")

    def execute(self, params: Dict) -> Dict:
        self._send_obj(params)
        return self._read_obj()    

    def disconnect(self):
        self.close()

    @staticmethod
    def command_template(command_name, arguments=None) -> Dict:
        return dict([('command', command_name), ('arguments', arguments or dict())])
        
    def command(self, command_name, arguments=None):
        return self.execute(self.command_template(command_name, arguments))

    def login(self, user_id, password, app_name='') -> str:
        arguments: Dict = dict(userId=user_id, password=password, appName=app_name)
        resp = self.command(command_name='login', arguments=arguments)
        if resp['status']:
            return resp['streamSessionId']
        else:
            logger.debug(f"Error {resp['errorCode']}, {resp['errorDescr']}")
            if not (resp['errorCode'] == 'BE118' and resp['errorDescr'] == 'User already logged'):
                raise LoginException(f"Error {resp['errorCode']}, {resp['errorDescr']}")

    def get_symbol(self, symbol: str) -> Dict:
        arguments: Dict = dict(symbol=symbol)
        resp = self.command(command_name="getSymbol", arguments=arguments)
        if resp['status']:
            return resp['returnData']
        else:
            logger.error(f"Error {resp['errorCode']}, {resp['errorDescr']}")
            raise GetSymbolException(f"Error {resp['errorCode']}, {resp['errorDescr']}")

    def get_trades(self, opened_only: bool = True) -> List:
        arguments: Dict = dict(openedOnly=opened_only)
        # try:
        resp = self.command(command_name="getTrades", arguments=arguments)
        if resp['status']:
            return resp['returnData']
        else:
            raise GetTradeException(f"Error {resp['errorCode']}, {resp['errorDescr']}")

    @retry(retry_on_exception=retry_if_get_tick_prices_exception,
           stop_max_attempt_number=5, wait_exponential_multiplier=1000, wait_exponential_max=10000)
    def get_tick_prices(self,
                        symbols: List[str], level: int = 0, timestamp: int = ts_to_unix_ts(datetime.now())) -> List:
        arguments: Dict = dict(symbols=symbols, level=level, timestamp=timestamp)
        resp = self.command(command_name="getTickPrices", arguments=arguments)
        if resp['status']:
            if resp['returnData']['quotations']:
                return resp['returnData']['quotations']
            else:
                raise GetTickPricesException(f"Error EMPTY, No tick prices received")
        raise GetTickPricesException(f"Error {resp['errorCode']}, {resp['errorDescr']}")

    def get_chart_range_request(
            self, period_value: int, symbol_name: str, start_unix_ts: int, end_unix_ts: int) -> List:
        arguments: Dict = dict(info=dict(
            symbol=symbol_name,
            period=period_value,
            start=start_unix_ts,
            end=end_unix_ts))
        resp = self.command(command_name="getChartRangeRequest", arguments=arguments)
        if resp['status']:
            return resp['returnData']['rateInfos']
        else:
            logger.error(f"Error {resp['errorCode']}:: {resp['errorDescr']}")
            return []
        
    def get_server_time(self) -> datetime:
        resp = self.command(command_name="getServerTime")
        if resp['status']:
            return unix_ts_to_ts(resp['returnData']['time'])
        else:
            logger.error(f"Error {resp['errorCode']}:: {resp['errorDescr']}")

    def get_trading_hours(self, symbols: List[str]) -> Dict[str, Dict[int, List[Dict[str, datetime]]]]:
        arguments: Dict = dict(symbols=symbols)
        resp = self.command(command_name="getTradingHours", arguments=arguments)
        if resp['status']:
            trading_hours = dict()
            for symbol in resp['returnData']:
                trading_hours[symbol['symbol']] = dict()
                for day in symbol['trading']:
                    if not day['day'] in trading_hours[symbol['symbol']]:
                        trading_hours[symbol['symbol']][day['day']] = []
                    trading_hours[symbol['symbol']][day['day']].append(
                        dict(from_ts=unix_t_to_ts(day['fromT']), to_ts=unix_t_to_ts(day['toT'])))
            return trading_hours
        else:
            logger.error(f"Error {resp['errorCode']}:: {resp['errorDescr']}")
            raise GetTradingHoursException(f"Error {resp['errorCode']}, {resp['errorDescr']}")

    def get_trade_records(self, orders: List[int]) -> List:
        arguments: Dict = dict(orders=orders)
        resp = self.command(command_name="getTradeRecords", arguments=arguments)
        if resp['status']:
            return resp['returnData']
        else:
            logger.error(f"Error {resp['errorCode']}:: {resp['errorDescr']}")
            raise GetTradeRecordsException(f"Error {resp['errorCode']}, {resp['errorDescr']}")

    def trade_transaction(self, trade: Dict, _type: int) -> int:
        keeping_keys = ['cmd', 'comment', 'expiration', 'offset', 'order', 'sl', 'symbol', 'tp', 'volume']
        arguments: Dict = dict(tradeTransInfo={key: trade[key] for key in keeping_keys})
        arguments['tradeTransInfo']['type'] = _type
        if _type == Type.OPEN:
            arguments['tradeTransInfo']['price'] = trade['open_price']
        elif _type == Type.CLOSE:
            arguments['tradeTransInfo']['price'] = trade['close_price']
        resp = self.command(command_name="tradeTransaction", arguments=arguments)
        if resp['status']:
            return resp['returnData']['order']
        else:
            logger.error(f"Error {resp['errorCode']}:: {resp['errorDescr']}")
            raise OrderNotSentException(f"Error {resp['errorCode']}:: {resp['errorDescr']}")

    def trade_transaction_status(self, transacting_order: int) -> Dict:
        arguments: Dict = dict(order=transacting_order)
        resp = self.command(command_name="tradeTransactionStatus", arguments=arguments)
        if resp['status']:
            return resp['returnData']
        else:
            logger.error(f"Error {resp['errorCode']}:: {resp['errorDescr']}")
            raise OrderStatusNotSentException(f"Error {resp['errorCode']}:: {resp['errorDescr']}")

    def get_margin_level(self) -> Dict:
        resp = self.command(command_name="getMarginLevel")
        if resp['status']:
            return resp['returnData']
        else:
            logger.error(f"Error {resp['errorCode']}, {resp['errorDescr']}")
            raise GetMarginLevel(f"Error {resp['errorCode']}, {resp['errorDescr']}")

    def get_free_margin(self) -> float:
        return self.get_margin_level()['margin_free']

    def get_margin_trade(self, symbol: str, volume: float) -> float:
        arguments: Dict = dict(symbol=symbol, volume=volume)
        resp = self.command(command_name="getMarginTrade", arguments=arguments)
        if resp['status']:
            return resp['returnData']['margin']
        else:
            logger.error(f"Error {resp['errorCode']}, {resp['errorDescr']}")
            raise GetMarginTrade(f"Error {resp['errorCode']}, {resp['errorDescr']}")
